{$N+}
{For Turbo Pascal 5.0, 6.0 or 7.0 by Han_kleijn@hnsky.org}
uses crt;
{=======================================================================}
type
  hnskyhdr = record
             nr1 : byte;
             nr2 : byte;
             nr3 : byte;
             ra1 : byte;
             ra2 : byte;
             ra3 : byte;
             dec1: byte;
             dec2: integer;
             mag2: shortint;
             spec2:  byte;
   end;
var
    Numwritten                  : Word;
    FromF                       : file;
    tof                         : file;
    maximum,step  :  real;
    ra_deg_start,ra_deg_stop  : integer;
    ra_rad_start,ra_rad_stop  : integer;
    ra_h_start,ra_h_stop      : integer;
    ra_m_start,ra_m_stop      : integer;
    ra_s_start,ra_s_stop      : integer;
    dec_sign                  : integer;
    dec_deg_start,dec_deg_stop: integer;
    dec_rad_start,dec_rad_stop: integer;
    dec_d_start,dec_d_stop    : integer;
    dec_m_start,dec_m_stop    : integer;
    dec_s_start,dec_s_stop    : integer;
    equinoxold,equinoxnew      : real;
    nr_start,nr_stop      : integer;
    mag_start,mag_stop    : integer;
    spec_start,spec_stop  : integer;
    accuracy              : char;
    infolength: integer;
    info      : array[1..512] of char;
    max,mag9,maxmag       : real;
    name,info1,info3      : string;
    error                 : integer;

    nr_of_stars           : longint;
    nr_of_skipped         : longint;
    eoffile               : boolean;
    er                    : text;

type
   real33= array[1..3,1..3] of real;
var
   X,Y,Z,teqxn,teqx,r,b,l  : real;
const
   testing:boolean=true;
(*-----------------------------------------------------------------------*)
(* SN: sine function (degrees)                                           *)
(*-----------------------------------------------------------------------*)
FUNCTION SN(X:REAL):REAL;
  CONST RAD=0.0174532925199433;
  BEGIN
    SN:=SIN(X*RAD)
  END;
(*-----------------------------------------------------------------------*)
(* CS: cosine function (degrees)                                         *)
(*-----------------------------------------------------------------------*)
FUNCTION CS(X:REAL):REAL;
  CONST RAD=0.0174532925199433;
  BEGIN
    CS:=COS(X*RAD)
  END;
(*-----------------------------------------------------------------------*)
(* ATN2: arctangent of y/x for two arguments                             *)
(*       (correct quadrant; -180 deg <= ATN2 <= +180 deg)                *)
(*-----------------------------------------------------------------------*)
FUNCTION ATN2(Y,X:REAL):REAL;
  CONST RAD=0.0174532925199433;
  VAR   AX,AY,PHI: REAL;
  BEGIN
    IF (X=0.0) AND (Y=0.0)
      THEN ATN2:=0.0
      ELSE
        BEGIN
          AX:=ABS(X); AY:=ABS(Y);
          IF (AX>AY)
            THEN PHI:=ARCTAN(AY/AX)/RAD
            ELSE PHI:=90.0-ARCTAN(AX/AY)/RAD;
          IF (X<0.0) THEN PHI:=180.0-PHI;
          IF (Y<0.0) THEN PHI:=-PHI;
          ATN2:=PHI;
        END;
  END;
  
(*-----------------------------------------------------------------------*)
(* POLAR: conversion of cartesian coordinates (x,y,z)                    *)
(*        into polar coordinates (r,theta,phi)                           *)
(*        (theta in [-90 deg,+90 deg]; phi in [0 deg,+360 deg])          *)
(*-----------------------------------------------------------------------*)
PROCEDURE POLAR(X,Y,Z:REAL;VAR R,THETA,PHI:REAL);
  VAR RHO: REAL;
  BEGIN
    RHO:=X*X+Y*Y;  R:=SQRT(RHO+Z*Z);
    PHI:=ATN2(Y,X); IF PHI<0 THEN PHI:=PHI+360.0;
    RHO:=SQRT(RHO); THETA:=ATN2(Z,RHO);
  END;

(*-----------------------------------------------------------------------*)
(* PRECART: calculate change of coordinates due to precession            *)
(*          for given transformation matrix A[i,j]                       *)
(*          (to be used with PMATECL und PMATEQU)                        *)
(*-----------------------------------------------------------------------*)
PROCEDURE PRECART(A:real33;VAR X,Y,Z:REAL);
  VAR U,V,W: REAL;
  BEGIN
    U := A[1,1]*X+A[1,2]*Y+A[1,3]*Z;
    V := A[2,1]*X+A[2,2]*Y+A[2,3]*Z;
    W := A[3,1]*X+A[3,2]*Y+A[3,3]*Z;
    X:=U; Y:=V; Z:=W;
  END;
(*-----------------------------------------------------------------------*)
(* CART: conversion of polar coordinates (r,theta,phi)                   *)
(*       into cartesian coordinates (x,y,z)                              *)
(*       (theta in [-90 deg,+90 deg]; phi in [-360 deg,+360 deg])        *)
(*-----------------------------------------------------------------------*)
PROCEDURE CART(R,THETA,PHI: REAL; VAR X,Y,Z: REAL);
  VAR RCST : REAL;
  BEGIN
    RCST := R*CS(THETA);
    X    := RCST*CS(PHI); Y := RCST*SN(PHI); Z := R*SN(THETA)
  END;

(*---------------------------------------------------------------------------*)
(* PMATEQU: Berechnung der Praezessionsmatrix A[i,j] fuer                    *)
(*          aequatoriale Koordinaten vom Aequinoktium T1 nach T2             *)
(*          ( T=(JD-2451545.0)/36525 )                                       *)
(*---------------------------------------------------------------------------*)
PROCEDURE PMATEQU(T1,T2:REAL; var a:real33);
  CONST SEC=3600.0;
  VAR DT,ZETA,Z,THETA: REAL;
      C1,S1,C2,S2,C3,S3: REAL;
  BEGIN
   DT:=T2-T1;
    ZETA  :=  ( (2306.2181+(1.39656-0.000139*T1)*T1)+
                ((0.30188-0.000345*T1)+0.017998*DT)*DT )*DT/SEC;
    Z     :=  ZETA + ( (0.79280+0.000411*T1)+0.000205*DT)*DT*DT/SEC;
    THETA :=  ( (2004.3109-(0.85330+0.000217*T1)*T1)-
                ((0.42665+0.000217*T1)+0.041833*DT)*DT )*DT/SEC;
    C1:=CS(Z);  C2:=CS(THETA);  C3:=CS(ZETA);
    S1:=SN(Z);  S2:=SN(THETA);  S3:=SN(ZETA);
    A[1,1]:=-S1*S3+C1*C2*C3; A[1,2]:=-S1*C3-C1*C2*S3; A[1,3]:=-C1*S2;
    A[2,1]:=+C1*S3+S1*C2*C3; A[2,2]:=+C1*C3-S1*C2*S3; A[2,3]:=-S1*S2;
    A[3,1]:=+S2*C3;          A[3,2]:=-S2*S3;          A[3,3]:=+C2;
  END;

PROCEDURE EP(yearold,yearnew,raOLD,decold:real;var ranew,decnew:real);
var teqxn,R : real;
    A       : REAL33;
begin
    CART(1,decold*180/pi,raold*180/pi,X,Y,Z);
    TEQX := (yearold-2000.0)/100.0;
    TEQXn := (yearnew-2000.0)/100.0;

    PMATEQU(TEQX,TEQXN,a);

    PRECART(A,X,Y,Z);

    TEQX := TEQXN;
    POLAR (X,Y,Z ,R,B,L);

    RAnew:=L*PI/180;
    DECnew:=B*PI/180;
end;

function  limit_radialen(z:real):real;
begin
  while z<0 do z:=z+2*pi;
  while z>2*pi do z:=z-2*pi;
  limit_radialen:=z;
end;
function remove_spaces(s:string):string;
var i:integer;
    s2:string;
begin
  s2:='';
  for i:=1 to length(s) do if s[i]<>' ' then s2:=s2+s[i];
  remove_spaces:=s2;
end;
FUNCTION SPECTRUM(ING:STRING):BYTE;
var
  out   :byte;
begin
 out:=0;
 case ing[1] of
             'O': out:=out+0*16;
             'B': out:=out+1*16;
             'A': out:=out+2*16;
             'F': out:=out+3*16;
             'G': out:=out+4*16;
             'K': out:=out+5*16;
             'M': out:=out+6*16;
             'R': out:=out+7*16;
             'N': out:=out+8*16;
             'S': out:=out+9*16;
             'C': out:=out+10*16;
             'W': out:=out+11*16;
             'P': out:=out+12*16;
{            '?': out:=out+13*16;}
             '+': out:=out+14*16;
             ' ': out:=out+15*16;
             else
               begin
                 out:=out+15*16; {blanco}
                 writeln('not found first spectrum letter:',ing[1]);
              {   halt;}
               end;
             end;{case}
 case ing[2] of
             '0': out:=out+0;
             '1': out:=out+1;
             '2': out:=out+2;
             '3': out:=out+3;
             '4': out:=out+4;
             '5': out:=out+5;
             '6': out:=out+6;
             '7': out:=out+7;
             '8': out:=out+8;
             '9': out:=out+9;
             'A': out:=out+10;
             'B': out:=out+11;
             'C': out:=out+12;
             'E': out:=out+13;
             '+': out:=out+14;
             ' ': out:=out+15;
             else
               begin
                  out:=out+15; {blanco}
{                 writeln('not found second spectrum letter :',ing[2]);}
{                 stop;      }
             end;
             end;
  spectrum:=out;
end;
procedure stopsort;
begin
 close(fromf);
 close(tof);
 halt;
end;
function copy_from_info(x,y:integer):string;
var i,j: integer;
begin

  if infolength<y then begin copy_from_info:='';exit; end;

  j:=1;
  for i:=x to y do
  begin
    copy_from_info[j]:=info[i];
    inc(j);
  end;
  copy_from_info[0]:=char(j-1);{set length}
end;
Procedure READVERYBIGASCII;{ 50 MBYTE DOES NOT WORK WITH READLN}
{version 00-03-27}
var
   numread,i  :word;

begin
   infolength:=0;
   BlockRead(FromF,info, SizeOf(info),NumRead);
   i:=0;
   repeat
     inc(i);
   until ((i>=numread) or (info[i]=#10));
   eoffile:=false;
   if info[i]=#10 then
   begin
     infolength:=i-1; {adjust length string}
     seek(fromf,filepos(fromf)-numread+infolength+1); {set pointer good}
   end
   else
   begin
     if SizeOf(info)=NumRead then
     begin
       writeln('Input file error');
       stopsort;
     end
     else
     eoffile:=true
   end; {no charactor 10 found}
end;

PROCEDURE sortpart(magn_start,step: real;writefil:boolean); {read file and find stars between magn_start and magn_start+step}
var info2:string;
    nr_of_errors          : longint;
    NumRead                     : Word;
    rax,decx,rax2,decx2         : real;
    magx                        : real;
    nr,rastart,decstart         : longint;
    hour,min                    : real;
    sec                         : real;
    buf2: array[1..12] of byte;

    p2       : ^hnskyhdr;		        { pointer to hnskyrecord }
    spec     :  string[3];
    I,step10,offset10 :  integer;

BEGIN
  filemode:=0;{otherwise reset does not accept files with read only attrib, this to prevent problems}
  Reset(FromF,1);
  { Open output file }
  nr_of_stars:=0;
  nr_of_skipped:=0;
  maxmag:=-2; {highest magnitude}
  step10:=round(step*10);{integer step in 0.1 magnitude}
  repeat
    nr_of_errors:=0;
    readverybigascii;
    if eoffile=true then
    begin
      close(fromf);
      exit;
    end;
    p2:= @buf2[1];			{ set pointer }
    with p2^ do
    begin
      info2:=copy_from_info(nr_start,nr_stop);
      for i:=1 to length(info2) do if info2[i]=' ' then info2[i]:='0';
      if copy_from_info(1,3)='BD-' then begin info2:='1'+info2; end;
      if copy_from_info(1,3)='SD-' then begin info2:='1'+info2; end;
      if copy_from_info(1,3)='CD-' then begin info2:='1'+info2;equinoxold:=1875;end
         else equinoxold:=1855;{for BD and SD}
      if copy_from_info(11,11)='a' then inc(nr_of_errors);;{export to error.txt file}
      if copy_from_info(11,11)='b' then inc(nr_of_errors);;
      if copy_from_info(11,11)='?' then inc(nr_of_errors);;
      if copy_from_info(11,11)='M' then inc(nr_of_errors);;
      if copy_from_info(11,11)='D' then inc(nr_of_errors);;


      info2:=remove_spaces(info2);
      val(info2,nr,error);
      if error<>0 then  begin  gotoxy(1,1);writeln('Nr error in: ',info2);inc(nr_of_errors);end;
      {make 3 byte word:}
      nr1 := lo(nr);{ important nr must be longint}
      nr2 := hi(nr);
      nr3 := nr shr 16;{=trunc(nr/(256*256))}
      {Right accension===========================================}

      if ra_deg_start<>0 then
      begin {Right accension degrees mode========}
        info2:=copy_from_info(ra_deg_start,ra_deg_stop);
        info2:=remove_spaces(info2);
        val(info2,rax,error);
        if error<>0 then begin
          gotoxy(1,1);writeln('RA_DEG error in:',info2);inc(nr_of_errors);end;
        rax:=rax*pi/180;
      end
      else
      if ra_rad_start<>0 then
      begin {Right accension radions mode=========}
        info2:=copy_from_info(ra_rad_start,ra_rad_stop);
        info2:=remove_spaces(info2);
        val(info2,rax,error);
        if error<>0 then begin
          gotoxy(1,1);writeln('RA_RAD error in:',info2);inc(nr_of_errors);;end;
      end
      else
      begin {Right accension seperate hr min sec mode========}
        info2:=copy_from_info(ra_h_start,ra_h_stop);
        info2:=remove_spaces(info2);
        val(info2,hour,error);
        if error<>0 then begin  gotoxy(1,1);
          writeln('RA error in:',info2);inc(nr_of_errors);;end;
        info2:=copy_from_info(ra_m_start,ra_m_stop);
        info2:=remove_spaces(info2);
        val(info2,min,error);
        if error<>0 then begin  gotoxy(1,1);
          writeln('RA error in:',info2);inc(nr_of_errors);;end;
        info2:=copy_from_info(ra_s_start,ra_s_stop);
        info2:=remove_spaces(info2);
        val(info2,sec,error);
        if error<>0 then begin  gotoxy(1,1);writeln('RA error in:',info2);inc(nr_of_errors);;end;
        rax:=(hour+min/60+sec/3600)*(2*pi/24);{radialen}
      end;


      {declination===========================================}
      decx:=0;
      if dec_rad_start<>0 then
      begin   {Declination radians mode========}
        info2:=copy_from_info(dec_rad_start,dec_rad_stop);
        info2:=remove_spaces(info2);
        val(info2,decx,error);
        if error<>0 then begin  gotoxy(1,1);
            writeln('DEC_RAD error in:',info2);inc(nr_of_errors);;end;
      end
      else
      if dec_d_start<>0 then
      begin   {Declination hr min sec mode========}
        info2:=copy_from_info(dec_d_start,dec_d_stop);
        info2:=remove_spaces(info2);
        val(info2,decx,error);
        if error<>0 then begin  gotoxy(1,1);writeln('DEC error in:',info2);inc(nr_of_errors);end;
        decx:=(decx)*(pi/180);
      end;
      if dec_m_start<>0 then
      begin
        info2:=copy_from_info(dec_m_start,dec_m_stop);
        info2:=remove_spaces(info2);
        val(info2,min,error);
        if error<>0 then begin  gotoxy(1,1);
           writeln('DEC error in:',info2);inc(nr_of_errors);end;
        decx:=decx+(min/60)*(pi/180);

      end;
      if dec_s_start<>0 then
      begin
        info2:=copy_from_info(dec_s_start,dec_s_stop);
        info2:=remove_spaces(info2);
        val(info2,sec,error);
        if error<>0 then begin  gotoxy(1,1);
          writeln('DEC error in:',info2);inc(nr_of_errors);;end;
        decx:=decx+(sec/3600)*(pi/180);
        end;
      if info[dec_sign]='-' then decx:=-decx;

      if equinoxold<>equinoxnew then EP(equinoxold,equinoxnew,rax,decx,rax2,decx2); {change equinox}

      rastart := round(((256*256*256)-1)*(limit_radialen(rax2))/(2*pi));
      {make 3 byte word:}
      ra1 := lo(rastart); {important rastart must be longint}
      ra2 := hi(rastart);
      ra3 := rastart shr 16;{make 3 byte word};


      decstart :=round(( ((256*256*128)-1)*(decx2)/(0.5*pi)));
      {make 3 byte integer:}
      dec2 := decstart shr 8 ;{make 3 byte integer, important decstart must be longint};
      dec1 := lo(decstart);


     {magnitude===========================================}
     info2:=copy_from_info(mag_start,mag_stop);
     info2:=remove_spaces(info2);
     val(info2,magx,error);
     if error<>0 then begin  gotoxy(1,1);writeln('Magn. error in:',info2);inc(nr_of_errors);end;
     if (((magx)>16 {changed for durchm was 25.4-2}) or (magx<-2.0)) then
       inc(nr_of_errors) {just don't know}
     else
     begin
       mag2:= round(magx*10);
                         {new magnitude 12.7 ==>  12.7}
                         {              12.8 ==> -12.8}
                         {              12.9 ==> -12.7}
                         {              13.0 ==> -12.6}
                         {              ..............}
                         {              23.4 ==> - 2.0 do not go further, confusion with Sirius=-1.7}
     end;

     {spectrum===========================================}
     if spec_start<>0 then
     begin
       spec:=copy_from_info(spec_start,spec_stop);
       spec2:=spectrum(spec);
     end
     else spec2:=spectrum('  ');

     if nr_of_errors=0 then
     begin


       offset10:=round((magx-magn_start)*10);{calculate offset integer value in 0.1 magnitude}
       if ((offset10<=step10) and (offset10>0)) then {within range, time to blockwrite.}
       { compare only integers otherwise compare errors or skipped stars due to tiny mathm errors !!!}
       begin
         if writefil then Blockwrite(tof,buf2, SizeOf(hnskyhdr),Numwritten)
         else
         begin if magx>maxmag then maxmag:=magx; {find highest magnitude} end;
         inc(nr_of_stars);
       end;
     end
     else
     begin
       inc(nr_of_skipped);
       if testing then {write errors only once}
          writeln(er,copy_from_info(1,infolength));
     end;
     end;
     until false;
END;

begin
  clrscr;
  if paramcount<1 then
  begin
    writeln('Program to make HNSKY star database from Durchmusterung files');
    writeln('I/122, I/119, I/114 ');
    writeln;
    writeln('Usage:  MAKE_BDM inputfile   (all data in one ASCII file.)');
    writeln;
    writeln('Equinox will be converted for BD and SD to from 1855 to 2000 and');
    writeln('for CD from 1875 to 2000');
    writeln;
    writeln('1) The input file should be clean ASCII. If you can strange errors or');
    writeln('   conversion does not progress, try to remove strange charactors in a word');
    writeln('   processor.');
    writeln('2) Stars with errors or with extension a, b will be exported to file error.txt');
    writeln('3) Stars with negative DEC will get a 1 in front');
    writeln('4) If the Bonner BD, Southern SD and Cordoba CD are combined then stars in');
    writeln('   the SD, area SD-1, SD-22  and SD-23 should be removed manually');
    writeln('   to a supplement since they get the same numbers as stars in BD-1, CD-22');
    writeln('   and CD-23 range.');
    writeln;
    writeln('Version 0.1 Han Kleijn');
    halt;
  end;
  write('Maximum magnitude ? '); readln(max);maximum:=round(max);
  writeln('Magnitude sorting accuracy normal=N (step 0.1), low=L(step 1.0)');
  write(  'or very low=V (step 5) ? [N, L or V] ');
      read(accuracy);if accuracy in ['n','N'] then step:=0.1 else
      if accuracy in ['l','L'] then step:=1 else step:=5; {step of magnitude}
    name:=paramstr(1);
  nr_start:=4;
  nr_stop:=10;

  ra_DEG_start:=0;
  ra_deg_stop:=0;
  ra_rad_start:=0;
  ra_rad_stop:=0;
  ra_h_start:=16;
  ra_h_stop:=17;
  ra_m_start:=18;
  ra_m_stop:=19;
  ra_s_start:=20;
  ra_s_stop:=23;
  dec_rad_start:=0;
  dec_rad_stop:=0;
  dec_sign:=24;
  dec_d_start:=25;
  dec_d_stop:=26;
  dec_m_start:=27;
  dec_m_stop:=30;
  dec_s_start:=0;
  dec_s_stop:=0;
  mag_start:=12;
  mag_stop:=15;
  spec_start:=0;
  spec_stop:=0;
  equinoxold:=1855;
  equinoxnew:=2000;
  {$I-}
  Assign(FromF,name);
  filemode:=0;{otherwise reset does not accept files with read only attrib, this to prevent problems}
  Reset(FromF,1);
  close(fromf);
  if ioresult<>0 then begin writeln('input file not found !!');stopsort;end;
  {$I-}

  assign(er,'error.txt');
  rewrite(er);

  writeln('Creating file ','BDM_HSKY.DAT');
  Assign(tof,'BDM_HSKY.DAT');
  { Record size = 1 }
  filemode:=2;{allow read/write}
  Rewrite(tof,1);{open output file}

  {make header begin===========================================================}
  clrscr;
  gotoxy(1,10);
  writeln('Testing nr of stars & maximum magnitude.');
  sortpart(-2,maximum- -2,false); {read file and find stars between magn_start and magn_start+step}
  if maxmag<maximum then maximum:=maxmag;{adjust maximum info}
  str(maximum:2:1,info1);
  str(nr_of_stars:0,info3);
  info1:='Durchmusterungen, Bonner, Southern, Cordoba' +
         ' '+info3+' stars, sorted to magnitude '+info1+'                                           '+
         '                                                                                                               ';
  info3:=info1;
  info1[0]:=#32; {first byte should be not the length but a nice space equals #32}
  testing:=false;
  Blockwrite(tof,info1,1+ 10*SizeOf(hnskyhdr),Numwritten);
  {make header end===========================================================}

  {sort file}
  mag9:=-2;{start magnitude}
  repeat
    clrscr;
    if mag9+step>maximum then step:=maximum-mag9;{adjust step before last go.}
    gotoxy(1,10);
    writeln('Header:   ',info3);
    writeln(nr_of_skipped,' stars will be skipped due to conversion errors.');
    writeln;
    writeln('Number of stars in last cycle: ',nr_of_stars);
    writeln;
    writeln('Now doing magnitude; ',mag9:3:1, ' to ',mag9+step:3:1);

    sortpart(mag9,step,true); {read file and find stars between magn_start and magn_start+step}
    mag9:=mag9+step;
  until mag9>=maximum;
  close(tof);
  close(er);
  gotoxy(1,18);

  writeln('Ready.')
end.
